# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import pytest

from pylsp import uris
from pylsp.plugins import signature
from pylsp.workspace import Document

DOC_URI = uris.from_fs_path(__file__)
DOC = """import sys

def main(param1, param2):
    \"\"\" Main docstring

    Args:
        param1 (str): Docs for param1
    \"\"\"
    raise Exception()

main(
"""

MULTI_LINE_DOC = """import sys

def main(param1=None,
            param2=None,
            param3=None,
            param4=None,
            param5=None,
            param6=None,
            param7=None,
            param8=None):
    \"\"\" Main docstring

    Args:
        param1 (str): Docs for param1
    \"\"\"
    raise Exception()

main(
"""


def test_no_signature(workspace):
    # Over blank line
    sig_position = {"line": 9, "character": 0}
    doc = Document(DOC_URI, workspace, DOC)

    sigs = signature.pylsp_signature_help(doc._config, doc, sig_position)["signatures"]
    assert not sigs


def test_signature(workspace):
    # Over '( ' in main(
    sig_position = {"line": 10, "character": 5}
    doc = Document(DOC_URI, workspace, DOC)

    sig_info = signature.pylsp_signature_help(doc._config, doc, sig_position)

    sigs = sig_info["signatures"]
    assert len(sigs) == 1
    assert sigs[0]["label"] == "main(param1, param2)"
    assert sigs[0]["parameters"][0]["label"] == "param1"
    assert sigs[0]["parameters"][0]["documentation"] == {
        "kind": "markdown",
        "value": "Docs for param1",
    }

    assert sig_info["activeParameter"] == 0


def test_multi_line_signature(workspace):
    # Over '( ' in main(
    sig_position = {"line": 17, "character": 5}
    doc = Document(DOC_URI, workspace, MULTI_LINE_DOC)

    sig_info = signature.pylsp_signature_help(doc._config, doc, sig_position)

    sigs = sig_info["signatures"]
    assert len(sigs) == 1
    assert sigs[0]["label"] == (
        "main(param1=None, param2=None, param3=None, param4=None, "
        "param5=None, param6=None, param7=None, param8=None)"
    )
    assert sigs[0]["parameters"][0]["label"] == "param1"
    assert sigs[0]["parameters"][0]["documentation"] == {
        "kind": "markdown",
        "value": "Docs for param1",
    }

    assert sig_info["activeParameter"] == 0


@pytest.mark.parametrize(
    "regex,doc",
    [
        (signature.SPHINX, "    :param test: parameter docstring"),
        (signature.EPYDOC, "    @param test: parameter docstring"),
        (signature.GOOGLE, "    test (str): parameter docstring"),
    ],
)
def test_docstring_params(regex, doc):
    m = regex.match(doc)
    assert m.group("param") == "test"
    assert m.group("doc") == "parameter docstring"



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import os

import pytest

from pylsp import uris
from pylsp.plugins.references import pylsp_references
from pylsp.workspace import Document

DOC1_NAME = "test1.py"
DOC2_NAME = "test2.py"

DOC1 = """class Test1():
    pass
"""

DOC2 = """from test1 import Test1

try:
    Test1()
except UnicodeError:
    pass
"""


@pytest.fixture
def tmp_workspace(temp_workspace_factory):
    return temp_workspace_factory(
        {
            DOC1_NAME: DOC1,
            DOC2_NAME: DOC2,
        }
    )


def test_references(tmp_workspace):
    # Over 'Test1' in class Test1():
    position = {"line": 0, "character": 8}
    DOC1_URI = uris.from_fs_path(os.path.join(tmp_workspace.root_path, DOC1_NAME))
    doc1 = Document(DOC1_URI, tmp_workspace)

    refs = pylsp_references(doc1, position, exclude_declaration=False)

    # Definition, the import and the instantiation
    assert len(refs) == 3

    # Briefly check excluding the definitions (also excludes imports, only counts uses)
    no_def_refs = pylsp_references(doc1, position, exclude_declaration=True)
    assert len(no_def_refs) == 1

    # Make sure our definition is correctly located
    doc1_ref = [u for u in refs if u["uri"] == DOC1_URI][0]
    assert doc1_ref["range"]["start"] == {"line": 0, "character": 6}
    assert doc1_ref["range"]["end"] == {"line": 0, "character": 11}

    # Make sure our import is correctly located
    doc2_import_ref = [u for u in refs if u["uri"] != DOC1_URI][0]
    assert doc2_import_ref["range"]["start"] == {"line": 0, "character": 18}
    assert doc2_import_ref["range"]["end"] == {"line": 0, "character": 23}

    doc2_usage_ref = [u for u in refs if u["uri"] != DOC1_URI][1]
    assert doc2_usage_ref["range"]["start"] == {"line": 3, "character": 4}
    assert doc2_usage_ref["range"]["end"] == {"line": 3, "character": 9}


def test_references_builtin(tmp_workspace):
    # Over 'UnicodeError':
    position = {"line": 4, "character": 7}
    doc2_uri = uris.from_fs_path(os.path.join(str(tmp_workspace.root_path), DOC2_NAME))
    doc2 = Document(doc2_uri, tmp_workspace)

    refs = pylsp_references(doc2, position, exclude_declaration=False)
    assert len(refs) >= 1

    expected = {
        "start": {"line": 4, "character": 7},
        "end": {"line": 4, "character": 19},
    }
    ranges = [r["range"] for r in refs]
    assert expected in ranges



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import math
import os
import sys
from pathlib import Path
from typing import Dict, NamedTuple

import pytest

from pylsp import lsp, uris
from pylsp._utils import JEDI_VERSION
from pylsp.plugins.jedi_completion import (
    pylsp_completion_item_resolve as pylsp_jedi_completion_item_resolve,
)
from pylsp.plugins.jedi_completion import pylsp_completions as pylsp_jedi_completions
from pylsp.plugins.rope_completion import pylsp_completions as pylsp_rope_completions
from pylsp.workspace import Document

PY2 = sys.version[0] == "2"
LINUX = sys.platform.startswith("linux")
CI = os.environ.get("CI")
LOCATION = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
DOC_URI = uris.from_fs_path(__file__)
DOC = """import os
print os.path.isabs("/tmp")

def hello():
    pass

def _a_hello():
    pass

class Hello():

    @property
    def world(self):
        return None

    def everyone(self, a, b, c=None, d=2):
        pass

print Hello().world

print Hello().every

def documented_hello():
    \"\"\"Sends a polite greeting\"\"\"
    pass
"""


def test_rope_import_completion(config, workspace):
    com_position = {"line": 0, "character": 7}
    doc = Document(DOC_URI, workspace, DOC)
    items = pylsp_rope_completions(config, workspace, doc, com_position)
    assert items is None


class TypeCase(NamedTuple):
    document: str
    position: dict
    label: str
    expected: lsp.CompletionItemKind


TYPE_CASES: Dict[str, TypeCase] = {
    "variable": TypeCase(
        document="test = 1\ntes",
        position={"line": 1, "character": 3},
        label="test",
        expected=lsp.CompletionItemKind.Variable,
    ),
    "function": TypeCase(
        document="def test():\n    pass\ntes",
        position={"line": 2, "character": 3},
        label="test()",
        expected=lsp.CompletionItemKind.Function,
    ),
    "keyword": TypeCase(
        document="fro",
        position={"line": 0, "character": 3},
        label="from",
        expected=lsp.CompletionItemKind.Keyword,
    ),
    "file": TypeCase(
        document='"' + __file__[:-2].replace('"', '\\"') + '"',
        position={"line": 0, "character": len(__file__) - 2},
        label=Path(__file__).name + '"',
        expected=lsp.CompletionItemKind.File,
    ),
    "module": TypeCase(
        document="import statis",
        position={"line": 0, "character": 13},
        label="statistics",
        expected=lsp.CompletionItemKind.Module,
    ),
    "class": TypeCase(
        document="KeyErr",
        position={"line": 0, "character": 6},
        label="KeyError",
        expected=lsp.CompletionItemKind.Class,
    ),
    "property": TypeCase(
        document=(
            "class A:\n"
            "    @property\n"
            "    def test(self):\n"
            "        pass\n"
            "A().tes"
        ),
        position={"line": 4, "character": 5},
        label="test",
        expected=lsp.CompletionItemKind.Property,
    ),
}


@pytest.mark.parametrize("case", list(TYPE_CASES.values()), ids=list(TYPE_CASES.keys()))
def test_jedi_completion_type(case, config, workspace):
    # property support was introduced in 0.18
    if case.expected == lsp.CompletionItemKind.Property and JEDI_VERSION.startswith(
        "0.17"
    ):
        return

    doc = Document(DOC_URI, workspace, case.document)
    items = pylsp_jedi_completions(config, doc, case.position)
    items = {i["label"]: i for i in items}
    assert items[case.label]["kind"] == case.expected


def test_jedi_completion(config, workspace):
    # Over 'i' in os.path.isabs(...)
    com_position = {"line": 1, "character": 15}
    doc = Document(DOC_URI, workspace, DOC)
    items = pylsp_jedi_completions(config, doc, com_position)

    assert items
    labels = [i["label"] for i in items]
    assert "isfile(path)" in labels

    # Test we don't throw with big character
    pylsp_jedi_completions(config, doc, {"line": 1, "character": 1000})


def test_jedi_completion_item_resolve(config, workspace):
    # Over the blank line
    com_position = {"line": 8, "character": 0}
    doc = Document(DOC_URI, workspace, DOC)
    config.update({"plugins": {"jedi_completion": {"resolve_at_most": math.inf}}})
    completions = pylsp_jedi_completions(config, doc, com_position)

    items = {c["label"]: c for c in completions}

    documented_hello_item = items["documented_hello()"]

    assert "documentation" not in documented_hello_item
    assert "detail" not in documented_hello_item

    resolved_documented_hello = pylsp_jedi_completion_item_resolve(
        doc._config, completion_item=documented_hello_item, document=doc
    )
    expected_doc = {
        "kind": "markdown",
        "value": "```python\ndocumented_hello()\n```\n\n\nSends a polite greeting",
    }
    assert resolved_documented_hello["documentation"] == expected_doc


def test_jedi_completion_with_fuzzy_enabled(config, workspace):
    # Over 'i' in os.path.isabs(...)
    config.update({"plugins": {"jedi_completion": {"fuzzy": True}}})
    com_position = {"line": 1, "character": 15}
    doc = Document(DOC_URI, workspace, DOC)

    items = pylsp_jedi_completions(config, doc, com_position)

    assert items

    expected = "commonprefix(m)"
    if JEDI_VERSION == "0.18.0":
        expected = "commonprefix(list)"
    assert items[0]["label"] == expected

    # Test we don't throw with big character
    pylsp_jedi_completions(config, doc, {"line": 1, "character": 1000})


def test_jedi_completion_resolve_at_most(config, workspace):
    # Over 'i' in os.path.isabs(...)
    com_position = {"line": 1, "character": 15}
    doc = Document(DOC_URI, workspace, DOC)

    # Do not resolve any labels
    config.update({"plugins": {"jedi_completion": {"resolve_at_most": 0}}})
    items = pylsp_jedi_completions(config, doc, com_position)
    labels = {i["label"] for i in items}
    assert "isabs" in labels

    # Resolve all items
    config.update({"plugins": {"jedi_completion": {"resolve_at_most": math.inf}}})
    items = pylsp_jedi_completions(config, doc, com_position)
    labels = {i["label"] for i in items}
    assert "isfile(path)" in labels


def test_rope_completion(config, workspace):
    # Over 'i' in os.path.isabs(...)
    com_position = {"line": 1, "character": 15}
    workspace.put_document(DOC_URI, source=DOC)
    doc = workspace.get_document(DOC_URI)
    items = pylsp_rope_completions(config, workspace, doc, com_position)

    assert items
    assert items[0]["label"] == "isabs"


def test_jedi_completion_ordering(config, workspace):
    # Over the blank line
    com_position = {"line": 8, "character": 0}
    doc = Document(DOC_URI, workspace, DOC)
    config.update({"plugins": {"jedi_completion": {"resolve_at_most": math.inf}}})
    completions = pylsp_jedi_completions(config, doc, com_position)

    items = {c["label"]: c["sortText"] for c in completions}

    # And that 'hidden' functions come after unhidden ones
    assert items["hello()"] < items["_a_hello()"]


def test_jedi_property_completion(config, workspace):
    # Over the 'w' in 'print Hello().world'
    com_position = {"line": 18, "character": 15}
    doc = Document(DOC_URI, workspace, DOC)
    completions = pylsp_jedi_completions(config, doc, com_position)

    items = {c["label"]: c["sortText"] for c in completions}

    # Ensure we can complete the 'world' property
    assert "world" in list(items.keys())[0]


def test_jedi_method_completion(config, workspace):
    # Over the 'y' in 'print Hello().every'
    com_position = {"line": 20, "character": 19}
    doc = Document(DOC_URI, workspace, DOC)

    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update({"plugins": {"jedi_completion": {"include_params": True}}})

    completions = pylsp_jedi_completions(config, doc, com_position)
    everyone_method = [
        completion
        for completion in completions
        if completion["label"] == "everyone(a, b, c, d)"
    ][0]

    # Ensure we only generate snippets for positional args
    assert everyone_method["insertTextFormat"] == lsp.InsertTextFormat.Snippet
    assert everyone_method["insertText"] == "everyone(${1:a}, ${2:b})$0"

    # Disable param snippets
    config.update({"plugins": {"jedi_completion": {"include_params": False}}})

    completions = pylsp_jedi_completions(config, doc, com_position)
    everyone_method = [
        completion
        for completion in completions
        if completion["label"] == "everyone(a, b, c, d)"
    ][0]

    assert "insertTextFormat" not in everyone_method
    assert everyone_method["insertText"] == "everyone"


@pytest.mark.skipif(
    PY2 or (sys.platform.startswith("linux") and os.environ.get("CI") is not None),
    reason="Test in Python 3 and not on CIs on Linux because wheels don't work on them.",
)
def test_pyqt_completion(config, workspace):
    # Over 'QA' in 'from PyQt5.QtWidgets import QApplication'
    doc_pyqt = "from PyQt5.QtWidgets import QA"
    com_position = {"line": 0, "character": len(doc_pyqt)}
    doc = Document(DOC_URI, workspace, doc_pyqt)
    completions = pylsp_jedi_completions(config, doc, com_position)

    assert completions is not None


def test_numpy_completions(config, workspace):
    doc_numpy = "import numpy as np; np."
    com_position = {"line": 0, "character": len(doc_numpy)}
    doc = Document(DOC_URI, workspace, doc_numpy)
    items = pylsp_jedi_completions(config, doc, com_position)

    assert items
    assert any("array" in i["label"] for i in items)


def test_pandas_completions(config, workspace):
    doc_pandas = "import pandas as pd; pd."
    com_position = {"line": 0, "character": len(doc_pandas)}
    doc = Document(DOC_URI, workspace, doc_pandas)
    items = pylsp_jedi_completions(config, doc, com_position)

    assert items
    assert any("DataFrame" in i["label"] for i in items)


def test_matplotlib_completions(config, workspace):
    doc_mpl = "import matplotlib.pyplot as plt; plt."
    com_position = {"line": 0, "character": len(doc_mpl)}
    doc = Document(DOC_URI, workspace, doc_mpl)
    items = pylsp_jedi_completions(config, doc, com_position)

    assert items
    assert any("plot" in i["label"] for i in items)


def test_snippets_completion(config, workspace):
    doc_snippets = "from collections import defaultdict \na=defaultdict"
    com_position = {"line": 0, "character": 35}
    doc = Document(DOC_URI, workspace, doc_snippets)
    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update({"plugins": {"jedi_completion": {"include_params": True}}})
    completions = pylsp_jedi_completions(config, doc, com_position)
    assert completions[0]["insertText"] == "defaultdict"

    com_position = {"line": 1, "character": len(doc_snippets)}
    completions = pylsp_jedi_completions(config, doc, com_position)
    assert completions[0]["insertText"] == "defaultdict($0)"
    assert completions[0]["insertTextFormat"] == lsp.InsertTextFormat.Snippet


def test_snippets_completion_at_most(config, workspace):
    doc_snippets = "from collections import defaultdict \na=defaultdict"
    doc = Document(DOC_URI, workspace, doc_snippets)
    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update({"plugins": {"jedi_completion": {"include_params": True}}})
    config.update({"plugins": {"jedi_completion": {"resolve_at_most": 0}}})

    com_position = {"line": 1, "character": len(doc_snippets)}
    completions = pylsp_jedi_completions(config, doc, com_position)
    assert completions[0]["insertText"] == "defaultdict"
    assert not completions[0].get("insertTextFormat", None)


def test_completion_with_class_objects(config, workspace):
    doc_text = "class FOOBAR(Object): pass\nFOOB"
    com_position = {"line": 1, "character": 4}
    doc = Document(DOC_URI, workspace, doc_text)
    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update(
        {
            "plugins": {
                "jedi_completion": {
                    "include_params": True,
                    "include_class_objects": True,
                }
            }
        }
    )
    completions = pylsp_jedi_completions(config, doc, com_position)
    assert len(completions) == 2

    assert completions[0]["label"] == "FOOBAR"
    assert completions[0]["kind"] == lsp.CompletionItemKind.Class

    assert completions[1]["label"] == "FOOBAR object"
    assert completions[1]["kind"] == lsp.CompletionItemKind.TypeParameter


def test_completion_with_function_objects(config, workspace):
    doc_text = "def foobar(): pass\nfoob"
    com_position = {"line": 1, "character": 4}
    doc = Document(DOC_URI, workspace, doc_text)
    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update(
        {
            "plugins": {
                "jedi_completion": {
                    "include_params": True,
                    "include_function_objects": True,
                }
            }
        }
    )
    completions = pylsp_jedi_completions(config, doc, com_position)
    assert len(completions) == 2

    assert completions[0]["label"] == "foobar()"
    assert completions[0]["kind"] == lsp.CompletionItemKind.Function

    assert completions[1]["label"] == "foobar() object"
    assert completions[1]["kind"] == lsp.CompletionItemKind.TypeParameter


def test_snippet_parsing(config, workspace):
    doc = "divmod"
    completion_position = {"line": 0, "character": 6}
    doc = Document(DOC_URI, workspace, doc)
    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update({"plugins": {"jedi_completion": {"include_params": True}}})
    completions = pylsp_jedi_completions(config, doc, completion_position)

    out = "divmod(${1:x}, ${2:y})$0"
    if JEDI_VERSION == "0.18.0":
        out = "divmod(${1:a}, ${2:b})$0"
    assert completions[0]["insertText"] == out


def test_multiline_import_snippets(config, workspace):
    document = "from datetime import(\n date,\n datetime)\na=date"
    doc = Document(DOC_URI, workspace, document)
    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update({"plugins": {"jedi_completion": {"include_params": True}}})

    position = {"line": 1, "character": 5}
    completions = pylsp_jedi_completions(config, doc, position)
    assert completions[0]["insertText"] == "date"

    position = {"line": 2, "character": 9}
    completions = pylsp_jedi_completions(config, doc, position)
    assert completions[0]["insertText"] == "datetime"


def test_multiline_snippets(config, workspace):
    document = "from datetime import\\\n date,\\\n datetime \na=date"
    doc = Document(DOC_URI, workspace, document)
    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update({"plugins": {"jedi_completion": {"include_params": True}}})

    position = {"line": 1, "character": 5}
    completions = pylsp_jedi_completions(config, doc, position)
    assert completions[0]["insertText"] == "date"

    position = {"line": 2, "character": 9}
    completions = pylsp_jedi_completions(config, doc, position)
    assert completions[0]["insertText"] == "datetime"


def test_multistatement_snippet(config, workspace):
    config.capabilities["textDocument"] = {
        "completion": {"completionItem": {"snippetSupport": True}}
    }
    config.update({"plugins": {"jedi_completion": {"include_params": True}}})

    document = "a = 1; from datetime import date"
    doc = Document(DOC_URI, workspace, document)
    position = {"line": 0, "character": len(document)}
    completions = pylsp_jedi_completions(config, doc, position)
    assert completions[0]["insertText"] == "date"

    document = "from math import fmod; a = fmod"
    doc = Document(DOC_URI, workspace, document)
    position = {"line": 0, "character": len(document)}
    completions = pylsp_jedi_completions(config, doc, position)
    assert completions[0]["insertText"] == "fmod(${1:x}, ${2:y})$0"


def test_jedi_completion_extra_paths(tmpdir, workspace):
    # Create a tempfile with some content and pass to extra_paths
    temp_doc_content = """
def spam():
    pass
"""
    p = tmpdir.mkdir("extra_path")
    extra_paths = [str(p)]
    p = p.join("foo.py")
    p.write(temp_doc_content)

    # Content of doc to test completion
    doc_content = """import foo
foo.s"""
    doc = Document(DOC_URI, workspace, doc_content)

    # After 'foo.s' without extra paths
    com_position = {"line": 1, "character": 5}
    completions = pylsp_jedi_completions(doc._config, doc, com_position)
    assert completions is None

    # Update config extra paths
    settings = {"pylsp": {"plugins": {"jedi": {"extra_paths": extra_paths}}}}
    doc.update_config(settings)

    # After 'foo.s' with extra paths
    com_position = {"line": 1, "character": 5}
    completions = pylsp_jedi_completions(doc._config, doc, com_position)
    assert completions[0]["label"] == "spam()"


@pytest.mark.skipif(
    PY2 or not LINUX or not CI, reason="tested on linux and python 3 only"
)
def test_jedi_completion_environment(workspace):
    # Content of doc to test completion
    doc_content = """import logh
"""
    doc = Document(DOC_URI, workspace, doc_content)

    # After 'import logh' with default environment
    com_position = {"line": 0, "character": 11}

    assert os.path.isdir("/tmp/pyenv/")

    settings = {"pylsp": {"plugins": {"jedi": {"environment": None}}}}
    doc.update_config(settings)
    completions = pylsp_jedi_completions(doc._config, doc, com_position)
    assert completions is None

    # Update config extra environment
    env_path = "/tmp/pyenv/bin/python"
    settings = {"pylsp": {"plugins": {"jedi": {"environment": env_path}}}}
    doc.update_config(settings)

    # After 'import logh' with new environment
    completions = pylsp_jedi_completions(doc._config, doc, com_position)
    assert completions[0]["label"] == "loghub"

    resolved = pylsp_jedi_completion_item_resolve(doc._config, completions[0], doc)
    assert "changelog generator" in resolved["documentation"]["value"].lower()


def test_document_path_completions(tmpdir, workspace_other_root_path):
    # Create a dummy module out of the workspace's root_path and try to get
    # completions for it in another file placed next to it.
    module_content = """
def foo():
    pass
"""

    p = tmpdir.join("mymodule.py")
    p.write(module_content)

    # Content of doc to test completion
    doc_content = """import mymodule
mymodule.f"""
    doc_path = str(tmpdir) + os.path.sep + "myfile.py"
    doc_uri = uris.from_fs_path(doc_path)
    doc = Document(doc_uri, workspace_other_root_path, doc_content)

    com_position = {"line": 1, "character": 10}
    completions = pylsp_jedi_completions(doc._config, doc, com_position)
    assert completions[0]["label"] == "foo()"


def test_file_completions(workspace, tmpdir):
    # Create directory and a file to get completions for them.
    # Note: `tmpdir`` is the root dir of the `workspace` fixture. That's why we use
    # it here.
    tmpdir.mkdir("bar")
    file = tmpdir.join("foo.txt")
    file.write("baz")

    # Content of doc to test completion
    doc_content = '"'
    doc = Document(DOC_URI, workspace, doc_content)

    # Request for completions
    com_position = {"line": 0, "character": 1}
    completions = pylsp_jedi_completions(doc._config, doc, com_position)

    # Check completions
    assert len(completions) == 2
    assert [c["kind"] == lsp.CompletionItemKind.File for c in completions]
    assert completions[0]["insertText"] == (
        ("bar" + "\\") if os.name == "nt" else ("bar" + "/")
    )
    assert completions[1]["insertText"] == 'foo.txt"'

    # When snippets are supported, ensure that path separators are escaped.
    support_snippet = {
        "textDocument": {"completion": {"completionItem": {"snippetSupport": True}}}
    }
    doc._config.capabilities.update(support_snippet)
    completions = pylsp_jedi_completions(doc._config, doc, com_position)
    assert completions[0]["insertText"] == (
        ("bar" + "\\\\") if os.name == "nt" else ("bar" + "\\/")
    )
    assert completions[1]["insertText"] == 'foo.txt"'



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import multiprocessing
import os
import sys
import time
from threading import Thread
from typing import Any, Dict, List
from unittest import mock

from docstring_to_markdown import UnknownFormatError
from flaky import flaky

from pylsp import _utils
from pylsp.lsp import NotebookCellKind
from pylsp.python_lsp import PythonLSPServer, start_io_lang_server

CALL_TIMEOUT_IN_SECONDS = 30


def send_notebook_did_open(client, cells: List[str]):
    """
    Sends a notebookDocument/didOpen notification with the given python cells.

    The notebook has the uri "notebook_uri" and the cells have the uris
    "cell_1_uri", "cell_2_uri", etc.
    """
    client._endpoint.notify(
        "notebookDocument/didOpen", notebook_with_python_cells(cells)
    )


def notebook_with_python_cells(cells: List[str]):
    """
    Create a notebook document with the given python cells.

    The notebook has the uri "notebook_uri" and the cells have the uris
    "cell_1_uri", "cell_2_uri", etc.
    """
    return {
        "notebookDocument": {
            "uri": "notebook_uri",
            "notebookType": "jupyter-notebook",
            "cells": [
                {
                    "kind": NotebookCellKind.Code,
                    "document": f"cell_{i+1}_uri",
                }
                for i in range(len(cells))
            ],
        },
        "cellTextDocuments": [
            {
                "uri": f"cell_{i+1}_uri",
                "languageId": "python",
                "text": cell,
            }
            for i, cell in enumerate(cells)
        ],
    }


def send_initialize_request(client, initialization_options: Dict[str, Any] = None):
    return client._endpoint.request(
        "initialize",
        {
            "processId": 1234,
            "rootPath": os.path.dirname(__file__),
            "initializationOptions": initialization_options,
        },
    ).result(timeout=CALL_TIMEOUT_IN_SECONDS)


def start(obj):
    obj.start()


class ClientServerPair:
    """
    A class to setup a client/server pair.

    args:
        start_server_in_process: if True, the server will be started in a process.
        check_parent_process: if True, the server_process will check if the parent process is alive.
    """

    def __init__(self, start_server_in_process=False, check_parent_process=False):
        # Client to Server pipe
        csr, csw = os.pipe()
        # Server to client pipe
        scr, scw = os.pipe()

        if start_server_in_process:
            ParallelKind = self._get_parallel_kind()
            self.server_process = ParallelKind(
                target=start_io_lang_server,
                args=(
                    os.fdopen(csr, "rb"),
                    os.fdopen(scw, "wb"),
                    check_parent_process,
                    PythonLSPServer,
                ),
            )
            self.server_process.start()
        else:
            self.server = PythonLSPServer(os.fdopen(csr, "rb"), os.fdopen(scw, "wb"))
            self.server_thread = Thread(target=start, args=[self.server])
            self.server_thread.start()

        self.client = PythonLSPServer(os.fdopen(scr, "rb"), os.fdopen(csw, "wb"))
        self.client_thread = Thread(target=start, args=[self.client])
        self.client_thread.start()

    def _get_parallel_kind(self):
        if os.name == "nt":
            return Thread

        if sys.version_info[:2] >= (3, 8):
            return multiprocessing.get_context("fork").Process

        return multiprocessing.Process


@flaky(max_runs=6, min_passes=1)
def test_debounce():
    interval = 0.1
    obj = mock.Mock()

    @_utils.debounce(0.1)
    def call_m():
        obj()

    assert not obj.mock_calls

    call_m()
    call_m()
    call_m()
    assert not obj.mock_calls

    time.sleep(interval * 2)
    assert len(obj.mock_calls) == 1

    call_m()
    time.sleep(interval * 2)
    assert len(obj.mock_calls) == 2


@flaky(max_runs=6, min_passes=1)
def test_debounce_keyed_by():
    interval = 0.1
    obj = mock.Mock()

    @_utils.debounce(0.1, keyed_by="key")
    def call_m(key):
        obj(key)

    assert not obj.mock_calls

    call_m(1)
    call_m(2)
    call_m(3)
    assert not obj.mock_calls

    time.sleep(interval * 2)
    obj.assert_has_calls(
        [
            mock.call(1),
            mock.call(2),
            mock.call(3),
        ],
        any_order=True,
    )
    assert len(obj.mock_calls) == 3

    call_m(1)
    call_m(1)
    call_m(1)
    time.sleep(interval * 2)
    assert len(obj.mock_calls) == 4


def test_list_to_string():
    assert _utils.list_to_string("string") == "string"
    assert _utils.list_to_string(["a", "r", "r", "a", "y"]) == "a,r,r,a,y"


def test_find_parents(tmpdir):
    subsubdir = tmpdir.ensure_dir("subdir", "subsubdir")
    path = subsubdir.ensure("path.py")
    test_cfg = tmpdir.ensure("test.cfg")

    assert _utils.find_parents(tmpdir.strpath, path.strpath, ["test.cfg"]) == [
        test_cfg.strpath
    ]


def test_merge_dicts():
    assert _utils.merge_dicts(
        {"a": True, "b": {"x": 123, "y": {"hello": "world"}}},
        {"a": False, "b": {"y": [], "z": 987}},
    ) == {"a": False, "b": {"x": 123, "y": [], "z": 987}}


def test_clip_column():
    assert _utils.clip_column(0, [], 0) == 0
    assert _utils.clip_column(2, ["123"], 0) == 2
    assert _utils.clip_column(3, ["123"], 0) == 3
    assert _utils.clip_column(5, ["123"], 0) == 3
    assert _utils.clip_column(0, ["\n", "123"], 0) == 0
    assert _utils.clip_column(1, ["\n", "123"], 0) == 0
    assert _utils.clip_column(2, ["123\n", "123"], 0) == 2
    assert _utils.clip_column(3, ["123\n", "123"], 0) == 3
    assert _utils.clip_column(4, ["123\n", "123"], 1) == 3


@mock.patch("docstring_to_markdown.convert")
def test_format_docstring_valid_rst_signature(mock_convert):
    """Test that a valid RST docstring includes the function signature."""
    docstring = """A function docstring.

    Parameters
    ----------
    a : str, something
    """

    # Mock the return value to avoid depedency on the real thing
    mock_convert.return_value = """A function docstring.

    #### Parameters

    - `a`: str, something
    """

    markdown = _utils.format_docstring(
        docstring,
        "markdown",
        ["something(a: str) -> str"],
    )["value"]

    assert markdown.startswith(
        _utils.wrap_signature("something(a: str) -> str"),
    )


@mock.patch("docstring_to_markdown.convert", side_effect=UnknownFormatError)
def test_format_docstring_invalid_rst_signature(_):
    """Test that an invalid RST docstring includes the function signature."""
    docstring = """A function docstring.

    Parameters
    ----------
    a : str, something
    """

    markdown = _utils.format_docstring(
        docstring,
        "markdown",
        ["something(a: str) -> str"],
    )["value"]

    assert markdown.startswith(
        _utils.wrap_signature("something(a: str) -> str"),
    )



# Copyright 2021- Python Language Server Contributors.

import time
from test.test_utils import (
    CALL_TIMEOUT_IN_SECONDS,
    send_initialize_request,
    send_notebook_did_open,
)
from unittest.mock import call, patch

import pytest

from pylsp import IS_WIN
from pylsp.lsp import NotebookCellKind
from pylsp.workspace import Notebook


def wait_for_condition(condition, timeout=CALL_TIMEOUT_IN_SECONDS):
    """Wait for a condition to be true, or timeout."""
    start_time = time.time()
    while not condition():
        time.sleep(0.1)
        if time.time() - start_time > timeout:
            raise TimeoutError("Timeout waiting for condition")


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_initialize(client_server_pair):
    client, server = client_server_pair
    response = send_initialize_request(client)
    assert server.workspace is not None
    selector = response["capabilities"]["notebookDocumentSync"]["notebookSelector"]
    assert isinstance(selector, list)


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_workspace_did_change_configuration(client_server_pair):
    """Test that we can update a workspace config w/o error when a notebook is open."""
    client, server = client_server_pair
    send_initialize_request(client)
    assert server.workspace is not None

    with patch.object(server._endpoint, "notify") as mock_notify:
        client._endpoint.notify(
            "notebookDocument/didOpen",
            {
                "notebookDocument": {
                    "uri": "notebook_uri",
                    "notebookType": "jupyter-notebook",
                    "cells": [
                        {
                            "kind": NotebookCellKind.Code,
                            "document": "cell_1_uri",
                        },
                    ],
                },
                "cellTextDocuments": [
                    {
                        "uri": "cell_1_uri",
                        "languageId": "python",
                        "text": "",
                    },
                ],
            },
        )
        wait_for_condition(lambda: mock_notify.call_count >= 1)
    assert isinstance(server.workspace.get_document("notebook_uri"), Notebook)
    assert len(server.workspace.documents) == 2

    server.workspace.update_config(
        {"pylsp": {"plugins": {"flake8": {"enabled": True}}}}
    )

    assert server.config.plugin_settings("flake8").get("enabled") is True
    assert (
        server.workspace.get_document("cell_1_uri")
        ._config.plugin_settings("flake8")
        .get("enabled")
        is True
    )


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_notebook_document__did_open(
    client_server_pair,
):
    client, server = client_server_pair
    send_initialize_request(client)

    with patch.object(server._endpoint, "notify") as mock_notify:
        # Test as many edge cases as possible for the diagnostics messages
        send_notebook_did_open(
            client, ["", "\n", "\nimport sys\n\nabc\n\n", "x", "y\n"]
        )
        wait_for_condition(lambda: mock_notify.call_count >= 5)
        expected_call_args = [
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_1_uri",
                    "diagnostics": [],
                },
            ),
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_2_uri",
                    "diagnostics": [],
                },
            ),
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_3_uri",
                    "diagnostics": [
                        {
                            "source": "pyflakes",
                            "range": {
                                "start": {"line": 1, "character": 0},
                                "end": {"line": 1, "character": 11},
                            },
                            "message": "'sys' imported but unused",
                            "severity": 2,
                        },
                        {
                            "source": "pyflakes",
                            "range": {
                                "start": {"line": 3, "character": 0},
                                "end": {"line": 3, "character": 4},
                            },
                            "message": "undefined name 'abc'",
                            "severity": 1,
                        },
                        {
                            "source": "pycodestyle",
                            "range": {
                                "start": {"line": 1, "character": 0},
                                "end": {"line": 1, "character": 11},
                            },
                            "message": "E303 too many blank lines (4)",
                            "code": "E303",
                            "severity": 2,
                        },
                    ],
                },
            ),
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_4_uri",
                    "diagnostics": [
                        {
                            "source": "pyflakes",
                            "range": {
                                "start": {"line": 0, "character": 0},
                                "end": {"line": 0, "character": 2},
                            },
                            "message": "undefined name 'x'",
                            "severity": 1,
                        },
                    ],
                },
            ),
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_5_uri",
                    "diagnostics": [
                        {
                            "source": "pyflakes",
                            "range": {
                                "start": {"line": 0, "character": 0},
                                "end": {"line": 0, "character": 2},
                            },
                            "message": "undefined name 'y'",
                            "severity": 1,
                        },
                    ],
                },
            ),
        ]
        mock_notify.assert_has_calls(expected_call_args)


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_notebook_document__did_change(
    client_server_pair,
):
    client, server = client_server_pair
    send_initialize_request(client)

    # Open notebook
    with patch.object(server._endpoint, "notify") as mock_notify:
        send_notebook_did_open(client, ["import sys", ""])
        wait_for_condition(lambda: mock_notify.call_count >= 2)
        assert len(server.workspace.documents) == 3
        for uri in ["cell_1_uri", "cell_2_uri", "notebook_uri"]:
            assert uri in server.workspace.documents
        assert len(server.workspace.get_document("notebook_uri").cells) == 2
        expected_call_args = [
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_1_uri",
                    "diagnostics": [
                        {
                            "source": "pyflakes",
                            "range": {
                                "start": {"line": 0, "character": 0},
                                "end": {"line": 0, "character": 11},
                            },
                            "message": "'sys' imported but unused",
                            "severity": 2,
                        }
                    ],
                },
            ),
            call(
                "textDocument/publishDiagnostics",
                params={"uri": "cell_2_uri", "diagnostics": []},
            ),
        ]
        mock_notify.assert_has_calls(expected_call_args)

    # Remove second cell
    with patch.object(server._endpoint, "notify") as mock_notify:
        client._endpoint.notify(
            "notebookDocument/didChange",
            {
                "notebookDocument": {
                    "uri": "notebook_uri",
                },
                "change": {
                    "cells": {
                        "structure": {
                            "array": {
                                "start": 1,
                                "deleteCount": 1,
                            },
                            "didClose": [
                                {
                                    "uri": "cell_2_uri",
                                }
                            ],
                        },
                    }
                },
            },
        )
        wait_for_condition(lambda: mock_notify.call_count >= 2)
        assert len(server.workspace.documents) == 2
        assert "cell_2_uri" not in server.workspace.documents
        assert len(server.workspace.get_document("notebook_uri").cells) == 1
        expected_call_args = [
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_1_uri",
                    "diagnostics": [
                        {
                            "source": "pyflakes",
                            "range": {
                                "start": {"line": 0, "character": 0},
                                "end": {"line": 0, "character": 10},
                            },
                            "message": "'sys' imported but unused",
                            "severity": 2,
                        },
                        {
                            "source": "pycodestyle",
                            "range": {
                                "start": {"line": 0, "character": 10},
                                "end": {"line": 0, "character": 10},
                            },
                            "message": "W292 no newline at end of file",
                            "code": "W292",
                            "severity": 2,
                        },
                    ],
                },
            )
        ]
        mock_notify.assert_has_calls(expected_call_args)

    # Add second cell
    with patch.object(server._endpoint, "notify") as mock_notify:
        client._endpoint.notify(
            "notebookDocument/didChange",
            {
                "notebookDocument": {
                    "uri": "notebook_uri",
                },
                "change": {
                    "cells": {
                        "structure": {
                            "array": {
                                "start": 1,
                                "deleteCount": 0,
                                "cells": [
                                    {
                                        "kind": NotebookCellKind.Code,
                                        "document": "cell_3_uri",
                                    }
                                ],
                            },
                            "didOpen": [
                                {
                                    "uri": "cell_3_uri",
                                    "languageId": "python",
                                    "text": "x",
                                }
                            ],
                        },
                    }
                },
            },
        )
        wait_for_condition(lambda: mock_notify.call_count >= 2)
        assert len(server.workspace.documents) == 3
        assert "cell_3_uri" in server.workspace.documents
        assert len(server.workspace.get_document("notebook_uri").cells) == 2
        expected_call_args = [
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_1_uri",
                    "diagnostics": [
                        {
                            "source": "pyflakes",
                            "range": {
                                "start": {"line": 0, "character": 0},
                                "end": {"line": 0, "character": 11},
                            },
                            "message": "'sys' imported but unused",
                            "severity": 2,
                        }
                    ],
                },
            ),
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_3_uri",
                    "diagnostics": [
                        {
                            "source": "pyflakes",
                            "range": {
                                "start": {"line": 0, "character": 0},
                                "end": {"line": 0, "character": 1},
                            },
                            "message": "undefined name 'x'",
                            "severity": 1,
                        },
                        {
                            "source": "pycodestyle",
                            "range": {
                                "start": {"line": 0, "character": 1},
                                "end": {"line": 0, "character": 1},
                            },
                            "message": "W292 no newline at end of file",
                            "code": "W292",
                            "severity": 2,
                        },
                    ],
                },
            ),
        ]
        mock_notify.assert_has_calls(expected_call_args)

    # Edit second cell
    with patch.object(server._endpoint, "notify") as mock_notify:
        client._endpoint.notify(
            "notebookDocument/didChange",
            {
                "notebookDocument": {
                    "uri": "notebook_uri",
                },
                "change": {
                    "cells": {
                        "textContent": [
                            {
                                "document": {
                                    "uri": "cell_3_uri",
                                },
                                "changes": [{"text": "sys.path"}],
                            }
                        ]
                    }
                },
            },
        )
        wait_for_condition(lambda: mock_notify.call_count >= 2)
        expected_call_args = [
            call(
                "textDocument/publishDiagnostics",
                params={"uri": "cell_1_uri", "diagnostics": []},
            ),
            call(
                "textDocument/publishDiagnostics",
                params={
                    "uri": "cell_3_uri",
                    "diagnostics": [
                        {
                            "source": "pycodestyle",
                            "range": {
                                "start": {"line": 0, "character": 8},
                                "end": {"line": 0, "character": 8},
                            },
                            "message": "W292 no newline at end of file",
                            "code": "W292",
                            "severity": 2,
                        }
                    ],
                },
            ),
        ]
        mock_notify.assert_has_calls(expected_call_args)


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_notebook__did_close(
    client_server_pair,
):
    client, server = client_server_pair
    send_initialize_request(client)

    # Open notebook
    with patch.object(server._endpoint, "notify") as mock_notify:
        send_notebook_did_open(client, ["import sys", ""])
        wait_for_condition(lambda: mock_notify.call_count >= 2)
        assert len(server.workspace.documents) == 3
        for uri in ["cell_1_uri", "cell_2_uri", "notebook_uri"]:
            assert uri in server.workspace.documents

    # Close notebook
    with patch.object(server._endpoint, "notify") as mock_notify:
        client._endpoint.notify(
            "notebookDocument/didClose",
            {
                "notebookDocument": {
                    "uri": "notebook_uri",
                },
                "cellTextDocuments": [
                    {
                        "uri": "cell_1_uri",
                    },
                    {
                        "uri": "cell_2_uri",
                    },
                ],
            },
        )
        wait_for_condition(lambda: mock_notify.call_count >= 2)
        assert len(server.workspace.documents) == 0


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_notebook_definition(client_server_pair):
    client, server = client_server_pair
    send_initialize_request(client)

    # Open notebook
    with patch.object(server._endpoint, "notify") as mock_notify:
        send_notebook_did_open(client, ["y=2\nx=1", "x"])
        # wait for expected diagnostics messages
        wait_for_condition(lambda: mock_notify.call_count >= 2)
        assert len(server.workspace.documents) == 3
        for uri in ["cell_1_uri", "cell_2_uri", "notebook_uri"]:
            assert uri in server.workspace.documents

    future = client._endpoint.request(
        "textDocument/definition",
        {
            "textDocument": {
                "uri": "cell_2_uri",
            },
            "position": {"line": 0, "character": 1},
        },
    )
    result = future.result(CALL_TIMEOUT_IN_SECONDS)
    assert result == [
        {
            "uri": "cell_1_uri",
            "range": {
                "start": {"line": 1, "character": 0},
                "end": {"line": 1, "character": 1},
            },
        }
    ]


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_notebook_completion(client_server_pair):
    """
    Tests that completions work across cell boundaries for notebook document support
    """
    client, server = client_server_pair
    send_initialize_request(client)

    # Open notebook
    with patch.object(server._endpoint, "notify") as mock_notify:
        send_notebook_did_open(
            client, ["answer_to_life_universe_everything = 42", "answer_"]
        )
        # wait for expected diagnostics messages
        wait_for_condition(lambda: mock_notify.call_count >= 2)
        assert len(server.workspace.documents) == 3
        for uri in ["cell_1_uri", "cell_2_uri", "notebook_uri"]:
            assert uri in server.workspace.documents

    future = client._endpoint.request(
        "textDocument/completion",
        {
            "textDocument": {
                "uri": "cell_2_uri",
            },
            "position": {"line": 0, "character": 7},
        },
    )
    result = future.result(CALL_TIMEOUT_IN_SECONDS)
    assert result == {
        "isIncomplete": False,
        "items": [
            {
                "data": {"doc_uri": "cell_2_uri"},
                "insertText": "answer_to_life_universe_everything",
                "kind": 6,
                "label": "answer_to_life_universe_everything",
                "sortText": "aanswer_to_life_universe_everything",
            },
        ],
    }



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import sys

from pylsp import lsp, uris
from pylsp.plugins import pyflakes_lint
from pylsp.workspace import Document

DOC_URI = uris.from_fs_path(__file__)
DOC = """import sys

def hello():
\tpass

import json
"""

DOC_SYNTAX_ERR = """def hello()
    pass
"""

DOC_UNDEFINED_NAME_ERR = "a = b"


DOC_ENCODING = """# encoding=utf-8
import sys
"""


def test_pyflakes(workspace):
    doc = Document(DOC_URI, workspace, DOC)
    diags = pyflakes_lint.pylsp_lint(workspace, doc)

    # One we're expecting is:
    msg = "'sys' imported but unused"
    unused_import = [d for d in diags if d["message"] == msg][0]

    assert unused_import["range"]["start"] == {"line": 0, "character": 0}
    assert unused_import["severity"] == lsp.DiagnosticSeverity.Warning


def test_syntax_error_pyflakes(workspace):
    doc = Document(DOC_URI, workspace, DOC_SYNTAX_ERR)
    diag = pyflakes_lint.pylsp_lint(workspace, doc)[0]

    if sys.version_info[:2] >= (3, 10):
        assert diag["message"] == "expected ':'"
    else:
        assert diag["message"] == "invalid syntax"
    assert diag["range"]["start"] == {"line": 0, "character": 12}
    assert diag["severity"] == lsp.DiagnosticSeverity.Error


def test_undefined_name_pyflakes(workspace):
    doc = Document(DOC_URI, workspace, DOC_UNDEFINED_NAME_ERR)
    diag = pyflakes_lint.pylsp_lint(workspace, doc)[0]

    assert diag["message"] == "undefined name 'b'"
    assert diag["range"]["start"] == {"line": 0, "character": 4}
    assert diag["severity"] == lsp.DiagnosticSeverity.Error


def test_unicode_encoding(workspace):
    doc = Document(DOC_URI, workspace, DOC_ENCODING)
    diags = pyflakes_lint.pylsp_lint(workspace, doc)

    assert len(diags) == 1
    assert diags[0]["message"] == "'sys' imported but unused"



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

from test.fixtures import DOC, DOC_URI

from pylsp.workspace import Document


def test_document_props(doc):
    assert doc.uri == DOC_URI
    assert doc.source == DOC


def test_document_lines(doc):
    assert len(doc.lines) == 4
    assert doc.lines[0] == "import sys\n"


def test_document_source_unicode(workspace):
    document_mem = Document(DOC_URI, workspace, "my source")
    document_disk = Document(DOC_URI, workspace)
    assert isinstance(document_mem.source, type(document_disk.source))


def test_offset_at_position(doc):
    assert doc.offset_at_position({"line": 0, "character": 8}) == 8
    assert doc.offset_at_position({"line": 1, "character": 5}) == 16
    assert doc.offset_at_position({"line": 2, "character": 0}) == 12
    assert doc.offset_at_position({"line": 2, "character": 4}) == 16
    assert doc.offset_at_position({"line": 4, "character": 0}) == 51


def test_word_at_position(doc):
    """Return the position under the cursor (or last in line if past the end)"""
    # import sys
    assert doc.word_at_position({"line": 0, "character": 8}) == "sys"
    # Past end of import sys
    assert doc.word_at_position({"line": 0, "character": 1000}) == "sys"
    # Empty line
    assert doc.word_at_position({"line": 1, "character": 5}) == ""
    # def main():
    assert doc.word_at_position({"line": 2, "character": 0}) == "def"
    # Past end of file
    assert doc.word_at_position({"line": 4, "character": 0}) == ""


def test_document_empty_edit(workspace):
    doc = Document("file:///uri", workspace, "")
    doc.apply_change(
        {
            "range": {
                "start": {"line": 0, "character": 0},
                "end": {"line": 0, "character": 0},
            },
            "text": "f",
        }
    )
    assert doc.source == "f"


def test_document_line_edit(workspace):
    doc = Document("file:///uri", workspace, "itshelloworld")
    doc.apply_change(
        {
            "text": "goodbye",
            "range": {
                "start": {"line": 0, "character": 3},
                "end": {"line": 0, "character": 8},
            },
        }
    )
    assert doc.source == "itsgoodbyeworld"


def test_document_multiline_edit(workspace):
    old = ["def hello(a, b):\n", "    print a\n", "    print b\n"]
    doc = Document("file:///uri", workspace, "".join(old))
    doc.apply_change(
        {
            "text": "print a, b",
            "range": {
                "start": {"line": 1, "character": 4},
                "end": {"line": 2, "character": 11},
            },
        }
    )
    assert doc.lines == ["def hello(a, b):\n", "    print a, b\n"]


def test_document_end_of_file_edit(workspace):
    old = ["print 'a'\n", "print 'b'\n"]
    doc = Document("file:///uri", workspace, "".join(old))
    doc.apply_change(
        {
            "text": "o",
            "range": {
                "start": {"line": 2, "character": 0},
                "end": {"line": 2, "character": 0},
            },
        }
    )
    assert doc.lines == [
        "print 'a'\n",
        "print 'b'\n",
        "o",
    ]



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

from test import unix_only, windows_only

import pytest

from pylsp import uris


@unix_only
@pytest.mark.parametrize(
    "uri,path",
    [
        ("file:///foo/bar#frag", "/foo/bar"),
        ("file:/foo/bar#frag", "/foo/bar"),
        ("file:/foo/space%20%3Fbar#frag", "/foo/space ?bar"),
    ],
)
def test_to_fs_path(uri, path):
    assert uris.to_fs_path(uri) == path


@windows_only
@pytest.mark.parametrize(
    "uri,path",
    [
        ("file:///c:/far/boo", "c:\\far\\boo"),
        ("file:///C:/far/boo", "c:\\far\\boo"),
        ("file:///C:/far/space%20%3Fboo", "c:\\far\\space ?boo"),
    ],
)
def test_win_to_fs_path(uri, path):
    assert uris.to_fs_path(uri) == path


@unix_only
@pytest.mark.parametrize(
    "path,uri",
    [
        ("/foo/bar", "file:///foo/bar"),
        ("/foo/space ?bar", "file:///foo/space%20%3Fbar"),
    ],
)
def test_from_fs_path(path, uri):
    assert uris.from_fs_path(path) == uri


@windows_only
@pytest.mark.parametrize(
    "path,uri",
    [
        ("c:\\far\\boo", "file:///c:/far/boo"),
        ("C:\\far\\space ?boo", "file:///c:/far/space%20%3Fboo"),
    ],
)
def test_win_from_fs_path(path, uri):
    assert uris.from_fs_path(path) == uri


@pytest.mark.parametrize(
    "uri,kwargs,new_uri",
    [
        ("file:///foo/bar", {"path": "/baz/boo"}, "file:///baz/boo"),
        (
            "file:///D:/hello%20world.py",
            {"path": "D:/hello universe.py"},
            "file:///d:/hello%20universe.py",
        ),
    ],
)
def test_uri_with(uri, kwargs, new_uri):
    assert uris.uri_with(uri, **kwargs) == new_uri



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import os

import pytest

from pylsp import uris
from pylsp.plugins.jedi_rename import pylsp_rename
from pylsp.workspace import Document

DOC_NAME = "test1.py"
DOC = """class Test1():
    pass

class Test2(Test1):
    pass
"""

DOC_NAME_EXTRA = "test2.py"
DOC_EXTRA = """from test1 import Test1
x = Test1()
"""

DOC_NAME_SIMPLE = "test3.py"
DOC_SIMPLE = "foo = 12"


@pytest.fixture
def tmp_workspace(temp_workspace_factory):
    return temp_workspace_factory(
        {DOC_NAME: DOC, DOC_NAME_EXTRA: DOC_EXTRA, DOC_NAME_SIMPLE: DOC_SIMPLE}
    )


def test_jedi_rename(tmp_workspace, config):
    # rename the `Test1` class
    position = {"line": 0, "character": 6}
    DOC_URI = uris.from_fs_path(os.path.join(tmp_workspace.root_path, DOC_NAME))
    doc = Document(DOC_URI, tmp_workspace)

    result = pylsp_rename(config, tmp_workspace, doc, position, "ShouldBeRenamed")
    assert len(result.keys()) == 1

    changes = result.get("documentChanges")
    assert len(changes) == 2

    assert changes[0]["textDocument"]["uri"] == doc.uri
    assert changes[0]["textDocument"]["version"] == doc.version
    assert changes[0].get("edits") == [
        {
            "range": {
                "start": {"line": 0, "character": 0},
                "end": {"line": 5, "character": 0},
            },
            "newText": "class ShouldBeRenamed():\n    pass\n\nclass Test2(ShouldBeRenamed):\n    pass\n",
        }
    ]

    path = os.path.join(tmp_workspace.root_path, DOC_NAME_EXTRA)
    uri_extra = uris.from_fs_path(path)
    assert changes[1]["textDocument"]["uri"] == uri_extra
    # This also checks whether documents not yet added via textDocument/didOpen
    # but that do need to be renamed in the project have a `null` version
    # number.
    assert changes[1]["textDocument"]["version"] is None

    expected = "from test1 import ShouldBeRenamed\nx = ShouldBeRenamed()\n"
    if os.name == "nt":
        # The .write method in the temp_workspace_factory functions writes
        # Windows-style line-endings.
        expected = expected.replace("\n", "\r\n")
    assert changes[1].get("edits") == [
        {
            "range": {
                "start": {"line": 0, "character": 0},
                "end": {"line": 2, "character": 0},
            },
            "newText": expected,
        }
    ]

    # Regression test for issue python-lsp/python-lsp-server#413
    # rename foo
    position = {"line": 0, "character": 0}
    DOC_URI = uris.from_fs_path(os.path.join(tmp_workspace.root_path, DOC_NAME_SIMPLE))
    doc = Document(DOC_URI, tmp_workspace)

    result = pylsp_rename(config, tmp_workspace, doc, position, "bar")
    assert len(result.keys()) == 1

    changes = result.get("documentChanges")
    assert len(changes) == 1

    assert changes[0]["textDocument"]["uri"] == doc.uri
    assert changes[0]["textDocument"]["version"] == doc.version
    assert changes[0].get("edits") == [
        {
            "range": {
                "start": {"line": 0, "character": 0},
                "end": {"line": 0, "character": 0},
            },
            "newText": "bar = 12",
        }
    ]



# Copyright 2017 Palantir Technologies, Inc.
import os
import pathlib

import pytest

from pylsp import uris

DOC_URI = uris.from_fs_path(__file__)
NOTEBOOK_URI = uris.from_fs_path("notebook_uri")


def path_as_uri(path):
    return pathlib.Path(os.path.abspath(path)).as_uri()


def test_local(pylsp):
    """Since the workspace points to the test directory"""
    assert pylsp.workspace.is_local()


def test_put_document(pylsp):
    pylsp.workspace.put_document(DOC_URI, "content")
    assert DOC_URI in pylsp.workspace._docs


def test_put_notebook_document(pylsp):
    pylsp.workspace.put_notebook_document(DOC_URI, "jupyter-notebook", [])
    assert DOC_URI in pylsp.workspace._docs


def test_put_cell_document(pylsp):
    pylsp.workspace.put_cell_document(DOC_URI, NOTEBOOK_URI, "python", "content")
    assert DOC_URI in pylsp.workspace._docs


def test_get_document(pylsp):
    pylsp.workspace.put_document(DOC_URI, "TEXT")
    assert pylsp.workspace.get_document(DOC_URI).source == "TEXT"


def test_get_missing_document(tmpdir, pylsp):
    source = "TEXT"
    doc_path = tmpdir.join("test_document.py")
    doc_path.write(source)
    doc_uri = uris.from_fs_path(str(doc_path))
    assert pylsp.workspace.get_document(doc_uri).source == "TEXT"


def test_rm_document(pylsp):
    pylsp.workspace.put_document(DOC_URI, "TEXT")
    assert pylsp.workspace.get_document(DOC_URI).source == "TEXT"
    pylsp.workspace.rm_document(DOC_URI)
    assert pylsp.workspace.get_document(DOC_URI)._source is None


@pytest.mark.parametrize(
    "metafiles", [("setup.py",), ("pyproject.toml",), ("setup.py", "pyproject.toml")]
)
def test_non_root_project(pylsp, metafiles):
    repo_root = os.path.join(pylsp.workspace.root_path, "repo-root")
    os.mkdir(repo_root)
    project_root = os.path.join(repo_root, "project-root")
    os.mkdir(project_root)

    for metafile in metafiles:
        with open(os.path.join(project_root, metafile), "w+", encoding="utf-8") as f:
            f.write("# " + metafile)

    test_uri = uris.from_fs_path(os.path.join(project_root, "hello/test.py"))
    pylsp.workspace.put_document(test_uri, "assert True")
    test_doc = pylsp.workspace.get_document(test_uri)
    assert project_root in test_doc.sys_path()


def test_root_project_with_no_setup_py(pylsp):
    """Default to workspace root."""
    workspace_root = pylsp.workspace.root_path
    test_uri = uris.from_fs_path(os.path.join(workspace_root, "hello/test.py"))
    pylsp.workspace.put_document(test_uri, "assert True")
    test_doc = pylsp.workspace.get_document(test_uri)
    assert workspace_root in test_doc.sys_path()


def test_multiple_workspaces_from_initialize(pylsp_w_workspace_folders):
    pylsp, workspace_folders = pylsp_w_workspace_folders

    assert len(pylsp.workspaces) == 2

    folders_uris = [uris.from_fs_path(str(folder)) for folder in workspace_folders]

    for folder_uri in folders_uris:
        assert folder_uri in pylsp.workspaces

    assert folders_uris[0] == pylsp.root_uri

    # Create file in the first workspace folder.
    file1 = workspace_folders[0].join("file1.py")
    file1.write("import os")
    msg1 = {"uri": path_as_uri(str(file1)), "version": 1, "text": "import os"}

    pylsp.m_text_document__did_open(textDocument=msg1)
    assert msg1["uri"] in pylsp.workspace._docs
    assert msg1["uri"] in pylsp.workspaces[folders_uris[0]]._docs

    # Create file in the second workspace folder.
    file2 = workspace_folders[1].join("file2.py")
    file2.write("import sys")
    msg2 = {"uri": path_as_uri(str(file2)), "version": 1, "text": "import sys"}

    pylsp.m_text_document__did_open(textDocument=msg2)
    assert msg2["uri"] not in pylsp.workspace._docs
    assert msg2["uri"] in pylsp.workspaces[folders_uris[1]]._docs


def test_multiple_workspaces(tmpdir, pylsp):
    workspace1_dir = tmpdir.mkdir("workspace1")
    workspace2_dir = tmpdir.mkdir("workspace2")
    file1 = workspace1_dir.join("file1.py")
    file2 = workspace2_dir.join("file1.py")
    file1.write("import os")
    file2.write("import sys")

    msg = {"uri": path_as_uri(str(file1)), "version": 1, "text": "import os"}

    pylsp.m_text_document__did_open(textDocument=msg)
    assert msg["uri"] in pylsp.workspace._docs

    added_workspaces = [
        {"uri": path_as_uri(str(x))} for x in (workspace1_dir, workspace2_dir)
    ]
    event = {"added": added_workspaces, "removed": []}
    pylsp.m_workspace__did_change_workspace_folders(event)

    for workspace in added_workspaces:
        assert workspace["uri"] in pylsp.workspaces

    workspace1_uri = added_workspaces[0]["uri"]
    assert msg["uri"] not in pylsp.workspace._docs
    assert msg["uri"] in pylsp.workspaces[workspace1_uri]._docs

    msg = {"uri": path_as_uri(str(file2)), "version": 1, "text": "import sys"}
    pylsp.m_text_document__did_open(textDocument=msg)

    workspace2_uri = added_workspaces[1]["uri"]
    assert msg["uri"] in pylsp.workspaces[workspace2_uri]._docs

    event = {"added": [], "removed": [added_workspaces[0]]}
    pylsp.m_workspace__did_change_workspace_folders(event)
    assert workspace1_uri not in pylsp.workspaces


def test_multiple_workspaces_wrong_removed_uri(pylsp, tmpdir):
    workspace = {"uri": str(tmpdir.mkdir("Test123"))}
    event = {"added": [], "removed": [workspace]}
    pylsp.m_workspace__did_change_workspace_folders(event)
    assert workspace["uri"] not in pylsp.workspaces


def test_root_workspace_changed(pylsp, tmpdir):
    test_uri = str(tmpdir.mkdir("Test123"))
    pylsp.root_uri = test_uri
    pylsp.workspace._root_uri = test_uri

    workspace1 = {"uri": test_uri}
    workspace2 = {"uri": str(tmpdir.mkdir("NewTest456"))}

    event = {"added": [workspace2], "removed": [workspace1]}
    pylsp.m_workspace__did_change_workspace_folders(event)

    assert workspace2["uri"] == pylsp.workspace._root_uri
    assert workspace2["uri"] == pylsp.root_uri


def test_root_workspace_not_changed(pylsp, tmpdir):
    # removed uri != root_uri
    test_uri_1 = str(tmpdir.mkdir("Test12"))
    pylsp.root_uri = test_uri_1
    pylsp.workspace._root_uri = test_uri_1
    workspace1 = {"uri": str(tmpdir.mkdir("Test1234"))}
    workspace2 = {"uri": str(tmpdir.mkdir("NewTest456"))}
    event = {"added": [workspace2], "removed": [workspace1]}
    pylsp.m_workspace__did_change_workspace_folders(event)
    assert test_uri_1 == pylsp.workspace._root_uri
    assert test_uri_1 == pylsp.root_uri
    # empty 'added' list
    test_uri_2 = str(tmpdir.mkdir("Test123"))
    new_root_uri = workspace2["uri"]
    pylsp.root_uri = test_uri_2
    pylsp.workspace._root_uri = test_uri_2
    workspace1 = {"uri": test_uri_2}
    event = {"added": [], "removed": [workspace1]}
    pylsp.m_workspace__did_change_workspace_folders(event)
    assert new_root_uri == pylsp.workspace._root_uri
    assert new_root_uri == pylsp.root_uri
    # empty 'removed' list
    event = {"added": [workspace1], "removed": []}
    pylsp.m_workspace__did_change_workspace_folders(event)
    assert new_root_uri == pylsp.workspace._root_uri
    assert new_root_uri == pylsp.root_uri
    # 'added' list has no 'uri'
    workspace2 = {"TESTuri": "Test1234"}
    event = {"added": [workspace2], "removed": [workspace1]}
    pylsp.m_workspace__did_change_workspace_folders(event)
    assert new_root_uri == pylsp.workspace._root_uri
    assert new_root_uri == pylsp.root_uri


def test_root_workspace_removed(tmpdir, pylsp):
    workspace1_dir = tmpdir.mkdir("workspace1")
    workspace2_dir = tmpdir.mkdir("workspace2")
    root_uri = pylsp.root_uri

    # Add workspaces to the pylsp
    added_workspaces = [
        {"uri": path_as_uri(str(x))} for x in (workspace1_dir, workspace2_dir)
    ]
    event = {"added": added_workspaces, "removed": []}
    pylsp.m_workspace__did_change_workspace_folders(event)

    # Remove the root workspace
    removed_workspaces = [{"uri": root_uri}]
    event = {"added": [], "removed": removed_workspaces}
    pylsp.m_workspace__did_change_workspace_folders(event)

    # Assert that the first of the workspaces (in alphabetical order) is now
    # the root workspace
    assert pylsp.root_uri == path_as_uri(str(workspace1_dir))
    assert pylsp.workspace._root_uri == path_as_uri(str(workspace1_dir))


@pytest.mark.skipif(os.name == "nt", reason="Fails on Windows")
def test_workspace_loads_pycodestyle_config(pylsp, tmpdir):
    workspace1_dir = tmpdir.mkdir("Test123")
    pylsp.root_uri = str(workspace1_dir)
    pylsp.workspace._root_uri = str(workspace1_dir)

    # Test that project settings are loaded
    workspace2_dir = tmpdir.mkdir("NewTest456")
    cfg = workspace2_dir.join("pycodestyle.cfg")
    cfg.write("[pycodestyle]\n" "max-line-length = 1000")

    workspace1 = {"uri": str(workspace1_dir)}
    workspace2 = {"uri": str(workspace2_dir)}

    event = {"added": [workspace2], "removed": [workspace1]}
    pylsp.m_workspace__did_change_workspace_folders(event)

    seetings = pylsp.workspaces[str(workspace2_dir)]._config.settings()
    assert seetings["plugins"]["pycodestyle"]["maxLineLength"] == 1000

    # Test that project settings prevail over server ones.
    server_settings = {"pylsp": {"plugins": {"pycodestyle": {"maxLineLength": 10}}}}
    pylsp.m_workspace__did_change_configuration(server_settings)
    assert seetings["plugins"]["pycodestyle"]["maxLineLength"] == 1000

    # Test switching to another workspace with different settings
    workspace3_dir = tmpdir.mkdir("NewTest789")
    cfg1 = workspace3_dir.join("pycodestyle.cfg")
    cfg1.write("[pycodestyle]\n" "max-line-length = 20")

    workspace3 = {"uri": str(workspace3_dir)}

    event = {"added": [workspace3], "removed": [workspace2]}
    pylsp.m_workspace__did_change_workspace_folders(event)

    seetings = pylsp.workspaces[str(workspace3_dir)]._config.settings()
    assert seetings["plugins"]["pycodestyle"]["maxLineLength"] == 20


def test_settings_of_added_workspace(pylsp, tmpdir):
    test_uri = str(tmpdir.mkdir("Test123"))
    pylsp.root_uri = test_uri
    pylsp.workspace._root_uri = test_uri

    # Set some settings for the server.
    server_settings = {
        "pylsp": {"plugins": {"jedi": {"environment": "/usr/bin/python3"}}}
    }
    pylsp.m_workspace__did_change_configuration(server_settings)

    # Create a new workspace.
    workspace1 = {"uri": str(tmpdir.mkdir("NewTest456"))}
    event = {"added": [workspace1]}
    pylsp.m_workspace__did_change_workspace_folders(event)

    # Assert settings are inherited from the server config.
    workspace1_object = pylsp.workspaces[workspace1["uri"]]
    workspace1_jedi_settings = workspace1_object._config.plugin_settings("jedi")
    assert workspace1_jedi_settings == server_settings["pylsp"]["plugins"]["jedi"]


def test_no_progress_without_capability(workspace, consumer):
    workspace._config.capabilities["window"] = {"workDoneProgress": False}

    with workspace.report_progress("some_title"):
        pass

    assert len(consumer.call_args_list) == 0


def test_progress_simple(workspace, consumer):
    workspace._config.capabilities["window"] = {"workDoneProgress": True}

    with workspace.report_progress("some_title"):
        pass

    init_call, *progress_calls = consumer.call_args_list

    assert init_call[0][0]["method"] == "window/workDoneProgress/create"

    # same method for all calls
    assert all(
        call[0][0]["method"] == "$/progress" for call in progress_calls
    ), consumer.call_args_list

    # same token used in all calls
    assert (
        len(
            {call[0][0]["params"]["token"] for call in progress_calls}
            | {init_call[0][0]["params"]["token"]}
        )
        == 1
    )

    assert [call[0][0]["params"]["value"] for call in progress_calls] == [
        {"kind": "begin", "title": "some_title"},
        {"kind": "end"},
    ]


@pytest.mark.parametrize("exc", [Exception("something"), TimeoutError()])
def test_progress_initialization_fails_but_is_skipped(
    workspace, consumer, endpoint, exc
):
    def failing_token_initialization(self, *_args, **_kwargs):
        raise exc

    endpoint._dispatcher.m_window__work_done_progress__create = (
        failing_token_initialization
    )

    workspace._config.capabilities["window"] = {"workDoneProgress": True}

    with workspace.report_progress("some_title", skip_token_initialization=True):
        pass

    # we only see the failing token initialization call, no other calls
    progress_calls = consumer.call_args_list
    assert all(call[0][0]["method"] == "$/progress" for call in progress_calls)
    assert len({call[0][0]["params"]["token"] for call in progress_calls}) == 1
    assert [call[0][0]["params"]["value"] for call in progress_calls] == [
        {"kind": "begin", "title": "some_title"},
        {"kind": "end"},
    ]


def test_progress_with_percent(workspace, consumer):
    workspace._config.capabilities["window"] = {"workDoneProgress": True}

    with workspace.report_progress(
        "some_title", "initial message", percentage=1
    ) as progress_message:
        progress_message("ten", 10)
        progress_message("fifty", 50)
        progress_message("ninety", 90)

    init_call, *progress_calls = consumer.call_args_list

    assert init_call[0][0]["method"] == "window/workDoneProgress/create"

    # same method for all progress calls
    assert all(call[0][0]["method"] == "$/progress" for call in progress_calls)

    # same token used in all calls
    assert (
        len(
            {call[0][0]["params"]["token"] for call in progress_calls}
            | {init_call[0][0]["params"]["token"]}
        )
        == 1
    )

    assert [call[0][0]["params"]["value"] for call in progress_calls] == [
        {
            "kind": "begin",
            "message": "initial message",
            "percentage": 1,
            "title": "some_title",
        },
        {"kind": "report", "message": "ten", "percentage": 10},
        {"kind": "report", "message": "fifty", "percentage": 50},
        {"kind": "report", "message": "ninety", "percentage": 90},
        {"kind": "end"},
    ]


def test_progress_with_exception(workspace, consumer):
    workspace._config.capabilities["window"] = {"workDoneProgress": True}

    class DummyError(Exception):
        pass

    try:
        with workspace.report_progress("some_title"):
            raise DummyError("something")
    except DummyError:
        # we're using a specific exception class here so
        # any other exceptions happening in progress
        # reporting would correctly be raised in the
        # test.
        pass

    init_call, *progress_calls = consumer.call_args_list
    assert init_call[0][0]["method"] == "window/workDoneProgress/create"

    # same method for all calls
    assert all(call[0][0]["method"] == "$/progress" for call in progress_calls)

    # same token used in all calls
    assert (
        len(
            {call[0][0]["params"]["token"] for call in progress_calls}
            | {init_call[0][0]["params"]["token"]}
        )
        == 1
    )

    assert [call[0][0]["params"]["value"] for call in progress_calls] == [
        {"kind": "begin", "title": "some_title"},
        {"kind": "end"},
    ]



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import os
import sys
import time
from test.test_utils import ClientServerPair, send_initialize_request

import pytest
from flaky import flaky
from pylsp_jsonrpc.exceptions import JsonRpcMethodNotFound

RUNNING_IN_CI = bool(os.environ.get("CI"))

CALL_TIMEOUT_IN_SECONDS = 10


@pytest.fixture
def client_exited_server():
    """A fixture that sets up a client/server pair that support checking parent process aliveness
    and assert the server has already exited
    """
    client_server_pair_obj = ClientServerPair(True, True)

    yield client_server_pair_obj

    assert client_server_pair_obj.server_process.is_alive() is False


@flaky(max_runs=10, min_passes=1)
@pytest.mark.skipif(sys.platform == "darwin", reason="Too flaky on Mac")
def test_initialize(client_server_pair):
    client, _ = client_server_pair
    response = client._endpoint.request(
        "initialize",
        {"rootPath": os.path.dirname(__file__), "initializationOptions": {}},
    ).result(timeout=CALL_TIMEOUT_IN_SECONDS)
    assert "capabilities" in response


@flaky(max_runs=10, min_passes=1)
@pytest.mark.skipif(
    not sys.platform.startswith("Linux"), reason="Skipped on win and flaky on mac"
)
def test_exit_with_parent_process_died(
    client_exited_server,
):
    # language server should have already exited before responding
    lsp_server, mock_process = (
        client_exited_server.client,
        client_exited_server.server_process,
    )
    # with pytest.raises(Exception):
    lsp_server._endpoint.request(
        "initialize",
        {
            "processId": mock_process.pid,
            "rootPath": os.path.dirname(__file__),
            "initializationOptions": {},
        },
    ).result(timeout=CALL_TIMEOUT_IN_SECONDS)

    mock_process.terminate()
    time.sleep(CALL_TIMEOUT_IN_SECONDS)
    assert not client_exited_server.client_thread.is_alive()


@flaky(max_runs=10, min_passes=1)
@pytest.mark.skipif(sys.platform.startswith("linux"), reason="Fails on linux")
def test_not_exit_without_check_parent_process_flag(
    client_server_pair,
):
    client, _ = client_server_pair
    response = send_initialize_request(client)
    assert "capabilities" in response


@flaky(max_runs=10, min_passes=1)
@pytest.mark.skipif(RUNNING_IN_CI, reason="This test is hanging on CI")
def test_missing_message(client_server_pair):
    client, _ = client_server_pair
    with pytest.raises(JsonRpcMethodNotFound):
        client._endpoint.request("unknown_method").result(
            timeout=CALL_TIMEOUT_IN_SECONDS
        )



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import sys
from textwrap import dedent

from pylsp import uris
from pylsp.plugins.folding import pylsp_folding_range
from pylsp.workspace import Document

DOC_URI = uris.from_fs_path(__file__)
DOC = dedent(
    """
def func(arg1, arg2, arg3,
         arg4, arg5, default=func(
             2, 3, 4
         )):
    return (2, 3,
            4, 5)

@decorator(
    param1,
    param2
)
def decorated_func(x, y, z):
    if x:
        return y
    elif y:
        return z
    elif x + y > z:
        return True
    else:
        return x

class A():
    def method(self, x1):
        def inner():
            return x1

        if x2:
            func(3, 4, 5, 6,
                 7)
        elif x3 < 2:
            pass
        else:
            more_complex_func(2, 3, 4, 5, 6,
                              8)
        return inner

a = 2
operation = (a_large_variable_that_fills_all_space +
             other_embarrasingly_long_variable - 2 * 3 / 5)

(a, b, c,
 d, e, f) = func(3, 4, 5, 6,
                 7, 8, 9, 10)

for i in range(0, 3):
    i += 1
    while x < i:
        expr = (2, 4)
        a = func(expr + i, arg2, arg3, arg4,
                 arg5, var(2, 3, 4,
                           5))
    for j in range(0, i):
        if i % 2 == 1:
            pass

compren = [x for x in range(0, 3)
           if x == 2]

with open('doc', 'r') as f:
    try:
        f / 0
    except:
        pass
    finally:
        raise SomeException()

def testC():
    pass
"""
)

SYNTAX_ERR = dedent(
    """
def func(arg1, arg2, arg3,
         arg4, arg5, default=func(
             2, 3, 4
         )):
    return (2, 3,
            4, 5)

class A(:
    pass

a = 2
operation = (a_large_variable_that_fills_all_space +
             other_embarrasingly_long_variable - 2 * 3 /

(a, b, c,
 d, e, f) = func(3, 4, 5, 6,
                 7, 8, 9, 10
a = 2
for i in range(0, 3)
    i += 1
    while x < i:
        expr = (2, 4)
        a = func(expr + i, arg2, arg3, arg4,
                 arg5, var(2, 3, 4,
                           5))
    for j in range(0, i):
        if i % 2 == 1:
            pass
"""
)


def test_folding(workspace):
    doc = Document(DOC_URI, workspace, DOC)
    ranges = pylsp_folding_range(doc)
    expected = [
        {"startLine": 1, "endLine": 6},
        {"startLine": 2, "endLine": 3},
        {"startLine": 5, "endLine": 6},
        {"startLine": 8, "endLine": 11},
        {"startLine": 12, "endLine": 20},
        {"startLine": 13, "endLine": 14},
        {"startLine": 15, "endLine": 16},
        {"startLine": 17, "endLine": 18},
        {"startLine": 19, "endLine": 20},
        {"startLine": 22, "endLine": 35},
        {"startLine": 23, "endLine": 35},
        {"startLine": 24, "endLine": 25},
        {"startLine": 27, "endLine": 29},
        {"startLine": 28, "endLine": 29},
        {"startLine": 30, "endLine": 31},
        {"startLine": 32, "endLine": 34},
        {"startLine": 33, "endLine": 34},
        {"startLine": 38, "endLine": 39},
        {"startLine": 41, "endLine": 43},
        {"startLine": 42, "endLine": 43},
        {"startLine": 45, "endLine": 54},
        {"startLine": 47, "endLine": 51},
        {"startLine": 49, "endLine": 51},
        {"startLine": 50, "endLine": 51},
        {"startLine": 52, "endLine": 54},
        {"startLine": 53, "endLine": 54},
        {"startLine": 56, "endLine": 57},
        {"startLine": 59, "endLine": 65},
        {"startLine": 60, "endLine": 61},
        {"startLine": 62, "endLine": 63},
        {"startLine": 64, "endLine": 65},
        {"startLine": 67, "endLine": 68},
    ]
    if sys.version_info[:2] >= (3, 9):
        # the argument list of the decorator is also folded in Python >= 3.9
        expected.insert(4, {"startLine": 9, "endLine": 10})

    assert ranges == expected


def test_folding_syntax_error(workspace):
    doc = Document(DOC_URI, workspace, SYNTAX_ERR)
    ranges = pylsp_folding_range(doc)
    expected = [
        {"startLine": 1, "endLine": 6},
        {"startLine": 2, "endLine": 3},
        {"startLine": 5, "endLine": 6},
        {"startLine": 8, "endLine": 9},
        {"startLine": 12, "endLine": 13},
        {"startLine": 15, "endLine": 17},
        {"startLine": 16, "endLine": 17},
        {"startLine": 19, "endLine": 28},
        {"startLine": 21, "endLine": 25},
        {"startLine": 23, "endLine": 25},
        {"startLine": 24, "endLine": 25},
        {"startLine": 26, "endLine": 28},
        {"startLine": 27, "endLine": 28},
    ]
    assert ranges == expected



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

from pylsp import lsp, uris
from pylsp.plugins.highlight import pylsp_document_highlight
from pylsp.workspace import Document

DOC_URI = uris.from_fs_path(__file__)
DOC = """a = "hello"
a.startswith("b")
"""


def test_highlight(workspace):
    # Over 'a' in a.startswith
    cursor_pos = {"line": 1, "character": 0}

    doc = Document(DOC_URI, workspace, DOC)
    assert pylsp_document_highlight(doc, cursor_pos) == [
        {
            "range": {
                "start": {"line": 0, "character": 0},
                "end": {"line": 0, "character": 1},
            },
            # The first usage is Write
            "kind": lsp.DocumentHighlightKind.Write,
        },
        {
            "range": {
                "start": {"line": 1, "character": 0},
                "end": {"line": 1, "character": 1},
            },
            # The second usage is Read
            "kind": lsp.DocumentHighlightKind.Read,
        },
    ]


SYS_DOC = """import sys
print sys.path
"""


def test_sys_highlight(workspace):
    cursor_pos = {"line": 0, "character": 8}

    doc = Document(DOC_URI, workspace, SYS_DOC)
    assert pylsp_document_highlight(doc, cursor_pos) == [
        {
            "range": {
                "start": {"line": 0, "character": 7},
                "end": {"line": 0, "character": 10},
            },
            "kind": lsp.DocumentHighlightKind.Write,
        },
        {
            "range": {
                "start": {"line": 1, "character": 6},
                "end": {"line": 1, "character": 9},
            },
            "kind": lsp.DocumentHighlightKind.Read,
        },
    ]



# Copyright 2022- Python Language Server Contributors.

from test.test_notebook_document import wait_for_condition
from test.test_utils import send_initialize_request, send_notebook_did_open
from typing import Any, Dict, List
from unittest.mock import Mock, patch

import jedi
import parso
import pytest

from pylsp import IS_WIN, lsp, uris
from pylsp.config.config import Config
from pylsp.plugins.rope_autoimport import (
    _get_score,
    _should_insert,
    cache,
    get_name_or_module,
    get_names,
)
from pylsp.plugins.rope_autoimport import (
    pylsp_completions as pylsp_autoimport_completions,
)
from pylsp.workspace import Workspace

DOC_URI = uris.from_fs_path(__file__)


def contains_autoimport_completion(suggestion: Dict[str, Any], module: str) -> bool:
    """Checks if `suggestion` contains an autoimport completion for `module`."""
    return suggestion.get("label", "") == module and "import" in suggestion.get(
        "detail", ""
    )


def contains_autoimport_quickfix(suggestion: Dict[str, Any], module: str) -> bool:
    """Checks if `suggestion` contains an autoimport quick fix for `module`."""
    return suggestion.get("title", "") == f"import {module}"


@pytest.fixture(scope="session")
def autoimport_workspace(tmp_path_factory) -> Workspace:
    "Special autoimport workspace. Persists across sessions to make in-memory sqlite3 database fast."
    workspace = Workspace(
        uris.from_fs_path(str(tmp_path_factory.mktemp("pylsp"))), Mock()
    )
    workspace._config = Config(workspace.root_uri, {}, 0, {})
    workspace._config.update(
        {
            "rope_autoimport": {
                "memory": True,
                "enabled": True,
                "completions": {"enabled": True},
                "code_actions": {"enabled": True},
            }
        }
    )
    cache.reload_cache(workspace._config, workspace, single_thread=True)
    yield workspace
    workspace.close()


@pytest.fixture
def completions(config: Config, autoimport_workspace: Workspace, request):
    document, position = request.param
    com_position = {"line": 0, "character": position}
    autoimport_workspace.put_document(DOC_URI, source=document)
    doc = autoimport_workspace.get_document(DOC_URI)
    yield pylsp_autoimport_completions(
        config, autoimport_workspace, doc, com_position, None
    )
    autoimport_workspace.rm_document(DOC_URI)


def should_insert(phrase: str, position: int):
    expr = parso.parse(phrase)
    word_node = expr.get_leaf_for_position((1, position))
    return _should_insert(expr, word_node)


def check_dict(query: Dict, results: List[Dict]) -> bool:
    for result in results:
        if all(result[key] == query[key] for key in query.keys()):
            return True
    return False


@pytest.mark.parametrize("completions", [("""pathli """, 6)], indirect=True)
def test_autoimport_completion(completions):
    assert completions
    assert check_dict(
        {"label": "pathlib", "kind": lsp.CompletionItemKind.Module}, completions
    )


@pytest.mark.parametrize("completions", [("""import """, 7)], indirect=True)
def test_autoimport_import(completions):
    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""pathlib""", 2)], indirect=True)
def test_autoimport_pathlib(completions):
    assert completions[0]["label"] == "pathlib"

    start = {"line": 0, "character": 0}
    edit_range = {"start": start, "end": start}
    assert completions[0]["additionalTextEdits"] == [
        {"range": edit_range, "newText": "import pathlib\n"}
    ]


@pytest.mark.parametrize("completions", [("""import test\n""", 10)], indirect=True)
def test_autoimport_import_with_name(completions):
    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""def func(s""", 10)], indirect=True)
def test_autoimport_function(completions):
    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""class Test""", 10)], indirect=True)
def test_autoimport_class(completions):
    assert len(completions) == 0


@pytest.mark.parametrize("completions", [("""\n""", 0)], indirect=True)
def test_autoimport_empty_line(completions):
    assert len(completions) == 0


@pytest.mark.parametrize(
    "completions", [("""class Test(NamedTupl):""", 20)], indirect=True
)
def test_autoimport_class_complete(completions):
    assert len(completions) > 0


@pytest.mark.parametrize(
    "completions", [("""class Test(NamedTupl""", 20)], indirect=True
)
def test_autoimport_class_incomplete(completions):
    assert len(completions) > 0


@pytest.mark.parametrize("completions", [("""def func(s:Lis""", 12)], indirect=True)
def test_autoimport_function_typing(completions):
    assert len(completions) > 0
    assert check_dict({"label": "List"}, completions)


@pytest.mark.parametrize(
    "completions", [("""def func(s : Lis ):""", 16)], indirect=True
)
def test_autoimport_function_typing_complete(completions):
    assert len(completions) > 0
    assert check_dict({"label": "List"}, completions)


@pytest.mark.parametrize(
    "completions", [("""def func(s : Lis ) -> Generat:""", 29)], indirect=True
)
def test_autoimport_function_typing_return(completions):
    assert len(completions) > 0
    assert check_dict({"label": "Generator"}, completions)


def test_autoimport_defined_name(config, workspace):
    document = """List = "hi"\nLis"""
    com_position = {"line": 1, "character": 3}
    workspace.put_document(DOC_URI, source=document)
    doc = workspace.get_document(DOC_URI)
    completions = pylsp_autoimport_completions(
        config, workspace, doc, com_position, None
    )
    workspace.rm_document(DOC_URI)
    assert not check_dict({"label": "List"}, completions)


class TestShouldInsert:
    def test_dot(self):
        assert not should_insert("""str.""", 4)

    def test_dot_partial(self):
        assert not should_insert("""str.metho\n""", 9)

    def test_comment(self):
        assert not should_insert("""#""", 1)

    def test_comment_indent(self):
        assert not should_insert("""    # """, 5)

    def test_from(self):
        assert not should_insert("""from """, 5)
        assert should_insert("""from """, 4)


def test_sort_sources():
    result1 = _get_score(1, "import pathlib", "pathlib", "pathli")
    result2 = _get_score(2, "import pathlib", "pathlib", "pathli")
    assert result1 < result2


def test_sort_statements():
    result1 = _get_score(
        2, "from importlib_metadata import pathlib", "pathlib", "pathli"
    )
    result2 = _get_score(2, "import pathlib", "pathlib", "pathli")
    assert result1 > result2


def test_sort_both():
    result1 = _get_score(
        3, "from importlib_metadata import pathlib", "pathlib", "pathli"
    )
    result2 = _get_score(2, "import pathlib", "pathlib", "pathli")
    assert result1 > result2


def test_get_names():
    source = """
    from a import s as e
    import blah, bleh
    hello = "str"
    a, b = 1, 2
    def someone():
        soemthing
    class sfa:
        sfiosifo
    """
    results = get_names(jedi.Script(code=source))
    assert results == {"blah", "bleh", "e", "hello", "someone", "sfa", "a", "b"}


# Tests ruff, flake8 and pyflakes messages
@pytest.mark.parametrize(
    "message",
    ["Undefined name `os`", "F821 undefined name 'numpy'", "undefined name 'numpy'"],
)
def test_autoimport_code_actions_get_correct_module_name(autoimport_workspace, message):
    source = "os.path.join('a', 'b')"
    autoimport_workspace.put_document(DOC_URI, source=source)
    doc = autoimport_workspace.get_document(DOC_URI)
    diagnostic = {
        "range": {
            "start": {"line": 0, "character": 0},
            "end": {"line": 0, "character": 2},
        },
        "message": message,
    }
    module_name = get_name_or_module(doc, diagnostic)
    autoimport_workspace.rm_document(DOC_URI)
    assert module_name == "os"


def make_context(module_name, line, character_start, character_end):
    return {
        "diagnostics": [
            {
                "message": f"undefined name '{module_name}'",
                "range": {
                    "start": {"line": line, "character": character_start},
                    "end": {"line": line, "character": character_end},
                },
            }
        ]
    }


def position(line, character):
    return {"line": line, "character": character}


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_autoimport_code_actions_and_completions_for_notebook_document(
    client_server_pair,
):
    client, server = client_server_pair
    send_initialize_request(
        client,
        {
            "pylsp": {
                "plugins": {
                    "rope_autoimport": {
                        "memory": True,
                        "enabled": True,
                        "completions": {"enabled": True},
                    },
                }
            }
        },
    )
    with patch.object(server._endpoint, "notify") as mock_notify:
        # Expectations:
        # 1. We receive an autoimport suggestion for "os" in the first cell because
        #    os is imported after that.
        # 2. We don't receive an autoimport suggestion for "os" in the second cell because it's
        #    already imported in the second cell.
        # 3. We don't receive an autoimport suggestion for "os" in the third cell because it's
        #    already imported in the second cell.
        # 4. We receive an autoimport suggestion for "sys" because it's not already imported.
        # 5. If diagnostics doesn't contain "undefined name ...", we send empty quick fix suggestions.
        send_notebook_did_open(client, ["os", "import os\nos", "os", "sys"])
        wait_for_condition(lambda: mock_notify.call_count >= 4)
        # We received diagnostics messages for every cell
        assert all(
            "textDocument/publishDiagnostics" in c.args
            for c in mock_notify.call_args_list
        )

    rope_autoimport_settings = server.workspace._config.plugin_settings(
        "rope_autoimport"
    )
    assert rope_autoimport_settings.get("completions", {}).get("enabled", False) is True
    assert rope_autoimport_settings.get("memory", False) is True
    wait_for_condition(lambda: not cache.thread.is_alive())

    # 1.
    quick_fixes = server.code_actions("cell_1_uri", {}, make_context("os", 0, 0, 2))
    assert any(s for s in quick_fixes if contains_autoimport_quickfix(s, "os"))

    completions = server.completions("cell_1_uri", position(0, 2)).get("items")
    assert any(s for s in completions if contains_autoimport_completion(s, "os"))

    # 2.
    # We don't test code actions here as in this case, there would be no code actions sent bc
    # there wouldn't be a diagnostics message.
    completions = server.completions("cell_2_uri", position(1, 2)).get("items")
    assert not any(s for s in completions if contains_autoimport_completion(s, "os"))

    # 3.
    # Same as in 2.
    completions = server.completions("cell_3_uri", position(0, 2)).get("items")
    assert not any(s for s in completions if contains_autoimport_completion(s, "os"))

    # 4.
    quick_fixes = server.code_actions("cell_4_uri", {}, make_context("sys", 0, 0, 3))
    assert any(s for s in quick_fixes if contains_autoimport_quickfix(s, "sys"))

    completions = server.completions("cell_4_uri", position(0, 3)).get("items")
    assert any(s for s in completions if contains_autoimport_completion(s, "sys"))

    # 5.
    context = {"diagnostics": [{"message": "A random message"}]}
    quick_fixes = server.code_actions("cell_4_uri", {}, context)
    assert len(quick_fixes) == 0



# Copyright 2021- Python Language Server Contributors.

from test.test_notebook_document import wait_for_condition
from test.test_utils import send_initialize_request
from unittest.mock import patch

import pytest

from pylsp import IS_WIN

INITIALIZATION_OPTIONS = {
    "pylsp": {
        "plugins": {
            "flake8": {"enabled": True},
            "pycodestyle": {"enabled": False},
            "pyflakes": {"enabled": False},
        },
    }
}


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_set_flake8_using_init_opts(client_server_pair):
    client, server = client_server_pair
    send_initialize_request(client, INITIALIZATION_OPTIONS)
    for key, value in INITIALIZATION_OPTIONS["pylsp"]["plugins"].items():
        assert server.workspace._config.settings().get("plugins").get(key).get(
            "enabled"
        ) == value.get("enabled")


@pytest.mark.skipif(IS_WIN, reason="Flaky on Windows")
def test_set_flake8_using_workspace_did_change_configuration(client_server_pair):
    client, server = client_server_pair
    send_initialize_request(client, None)
    assert (
        server.workspace._config.settings().get("plugins").get("flake8").get("enabled")
        is False
    )

    with patch.object(server.workspace, "update_config") as mock_update_config:
        client._endpoint.notify(
            "workspace/didChangeConfiguration",
            {"settings": INITIALIZATION_OPTIONS},
        )
        wait_for_condition(lambda: mock_update_config.call_count >= 1)

        for key, value in INITIALIZATION_OPTIONS["pylsp"]["plugins"].items():
            assert server.workspace._config.settings().get("plugins").get(key).get(
                "enabled"
            ) == value.get("enabled")



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import os

from pylsp import uris
from pylsp.plugins.definition import pylsp_definitions
from pylsp.workspace import Document

DOC_URI = uris.from_fs_path(__file__)
DOC = """def a():
    pass

print(a())


class Directory(object):
    def __init__(self):
        self.members = dict()

    def add_member(self, id, name):
        self.members[id] = name


subscripted_before_reference = {}
subscripted_before_reference[0] = 0
subscripted_before_reference


def my_func():
    print('called')

alias = my_func
my_list = [1, None, alias]
inception = my_list[2]

inception()

import numpy
numpy.ones
"""


def test_definitions(config, workspace):
    # Over 'a' in print a
    cursor_pos = {"line": 3, "character": 6}

    # The definition of 'a'
    def_range = {
        "start": {"line": 0, "character": 4},
        "end": {"line": 0, "character": 5},
    }

    doc = Document(DOC_URI, workspace, DOC)
    assert [{"uri": DOC_URI, "range": def_range}] == pylsp_definitions(
        config, doc, cursor_pos
    )


def test_indirect_definitions(config, workspace):
    # Over 'subscripted_before_reference'
    cursor_pos = {"line": 16, "character": 0}

    # The definition of 'subscripted_before_reference',
    # skipping intermediate writes to the most recent definition
    def_range = {
        "start": {"line": 14, "character": 0},
        "end": {"line": 14, "character": len("subscripted_before_reference")},
    }

    doc = Document(DOC_URI, workspace, DOC)
    assert [{"uri": DOC_URI, "range": def_range}] == pylsp_definitions(
        config, doc, cursor_pos
    )


def test_definition_with_multihop_inference_goto(config, workspace):
    # Over 'inception()'
    cursor_pos = {"line": 26, "character": 0}

    # The most recent definition of 'inception',
    # ignoring alias hops
    def_range = {
        "start": {"line": 24, "character": 0},
        "end": {"line": 24, "character": len("inception")},
    }

    doc = Document(DOC_URI, workspace, DOC)
    assert [{"uri": DOC_URI, "range": def_range}] == pylsp_definitions(
        config, doc, cursor_pos
    )


def test_numpy_definition(config, workspace):
    # Over numpy.ones
    cursor_pos = {"line": 29, "character": 8}

    doc = Document(DOC_URI, workspace, DOC)
    defns = pylsp_definitions(config, doc, cursor_pos)
    assert len(defns) > 0, defns


def test_builtin_definition(config, workspace):
    # Over 'i' in dict
    cursor_pos = {"line": 8, "character": 24}

    doc = Document(DOC_URI, workspace, DOC)
    orig_settings = config.settings()

    # Check definition for `dict` goes to `builtins.pyi::dict`
    follow_defns_setting = {"follow_builtin_definitions": True}
    settings = {"plugins": {"jedi_definition": follow_defns_setting}}
    config.update(settings)
    defns = pylsp_definitions(config, doc, cursor_pos)
    assert len(defns) == 1
    assert defns[0]["uri"].endswith("builtins.pyi")

    # Check no definitions for `dict`
    follow_defns_setting["follow_builtin_definitions"] = False
    config.update(settings)
    defns = pylsp_definitions(config, doc, cursor_pos)
    assert not defns

    config.update(orig_settings)


def test_assignment(config, workspace):
    # Over 's' in self.members[id]
    cursor_pos = {"line": 11, "character": 19}

    # The assignment of 'self.members'
    def_range = {
        "start": {"line": 8, "character": 13},
        "end": {"line": 8, "character": 20},
    }

    doc = Document(DOC_URI, workspace, DOC)
    assert [{"uri": DOC_URI, "range": def_range}] == pylsp_definitions(
        config, doc, cursor_pos
    )


def test_document_path_definitions(config, workspace_other_root_path, tmpdir):
    # Create a dummy module out of the workspace's root_path and try to get
    # a definition on it in another file placed next to it.
    module_content = """
def foo():
    pass
"""

    p = tmpdir.join("mymodule.py")
    p.write(module_content)

    # Content of doc to test definition
    doc_content = """from mymodule import foo"""
    doc_path = str(tmpdir) + os.path.sep + "myfile.py"
    doc_uri = uris.from_fs_path(doc_path)
    doc = Document(doc_uri, workspace_other_root_path, doc_content)

    # The range where is defined in mymodule.py
    def_range = {
        "start": {"line": 1, "character": 4},
        "end": {"line": 1, "character": 7},
    }

    # The position where foo is called in myfile.py
    cursor_pos = {"line": 0, "character": 24}

    # The uri for mymodule.py
    module_path = str(p)
    module_uri = uris.from_fs_path(module_path)

    assert [{"uri": module_uri, "range": def_range}] == pylsp_definitions(
        config, doc, cursor_pos
    )



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

from pylsp import uris
from pylsp.text_edit import OverLappingTextEditException, apply_text_edits

DOC_URI = uris.from_fs_path(__file__)


def test_apply_text_edits_insert(pylsp):
    pylsp.workspace.put_document(DOC_URI, "012345678901234567890123456789")
    test_doc = pylsp.workspace.get_document(DOC_URI)

    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 0},
                        "end": {"line": 0, "character": 0},
                    },
                    "newText": "Hello",
                }
            ],
        )
        == "Hello012345678901234567890123456789"
    )
    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 1},
                        "end": {"line": 0, "character": 1},
                    },
                    "newText": "Hello",
                }
            ],
        )
        == "0Hello12345678901234567890123456789"
    )
    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 1},
                        "end": {"line": 0, "character": 1},
                    },
                    "newText": "Hello",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 1},
                        "end": {"line": 0, "character": 1},
                    },
                    "newText": "World",
                },
            ],
        )
        == "0HelloWorld12345678901234567890123456789"
    )
    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 2},
                        "end": {"line": 0, "character": 2},
                    },
                    "newText": "One",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 1},
                        "end": {"line": 0, "character": 1},
                    },
                    "newText": "Hello",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 1},
                        "end": {"line": 0, "character": 1},
                    },
                    "newText": "World",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 2},
                        "end": {"line": 0, "character": 2},
                    },
                    "newText": "Two",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 2},
                        "end": {"line": 0, "character": 2},
                    },
                    "newText": "Three",
                },
            ],
        )
        == "0HelloWorld1OneTwoThree2345678901234567890123456789"
    )


def test_apply_text_edits_replace(pylsp):
    pylsp.workspace.put_document(DOC_URI, "012345678901234567890123456789")
    test_doc = pylsp.workspace.get_document(DOC_URI)

    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "Hello",
                }
            ],
        )
        == "012Hello678901234567890123456789"
    )
    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "Hello",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 6},
                        "end": {"line": 0, "character": 9},
                    },
                    "newText": "World",
                },
            ],
        )
        == "012HelloWorld901234567890123456789"
    )
    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "Hello",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 6},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "World",
                },
            ],
        )
        == "012HelloWorld678901234567890123456789"
    )
    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 6},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "World",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "Hello",
                },
            ],
        )
        == "012HelloWorld678901234567890123456789"
    )
    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 3},
                    },
                    "newText": "World",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "Hello",
                },
            ],
        )
        == "012WorldHello678901234567890123456789"
    )


def test_apply_text_edits_overlap(pylsp):
    pylsp.workspace.put_document(DOC_URI, "012345678901234567890123456789")
    test_doc = pylsp.workspace.get_document(DOC_URI)

    did_throw = False
    try:
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "Hello",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 3},
                    },
                    "newText": "World",
                },
            ],
        )
    except OverLappingTextEditException:
        did_throw = True

    assert did_throw

    did_throw = False

    try:
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 0, "character": 3},
                        "end": {"line": 0, "character": 6},
                    },
                    "newText": "Hello",
                },
                {
                    "range": {
                        "start": {"line": 0, "character": 4},
                        "end": {"line": 0, "character": 4},
                    },
                    "newText": "World",
                },
            ],
        )
    except OverLappingTextEditException:
        did_throw = True

    assert did_throw


def test_apply_text_edits_multiline(pylsp):
    pylsp.workspace.put_document(DOC_URI, "0\n1\n2\n3\n4")
    test_doc = pylsp.workspace.get_document(DOC_URI)

    assert (
        apply_text_edits(
            test_doc,
            [
                {
                    "range": {
                        "start": {"line": 2, "character": 0},
                        "end": {"line": 3, "character": 0},
                    },
                    "newText": "Hello",
                },
                {
                    "range": {
                        "start": {"line": 1, "character": 1},
                        "end": {"line": 1, "character": 1},
                    },
                    "newText": "World",
                },
            ],
        )
        == "0\n1World\nHello3\n4"
    )
